# Copyright 2018, 2019 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
from __future__ import division

import time
import json
import select
import random
from netq_lib.common.enums import ConfKw
from netq_agent.cmd.netq_agent_discovery import NetQAgentDiscovery, AGENTSTATUS, ChannelState
from netq_lib.common.netq_pq import QueueElem

from concurrent import futures

# GRPC Specfic Imports
import grpc
from netq_agent.interface.netq_pb2 import (
    NetqAgentMessage,
)
from netq_agent.interface.netq_pb2_grpc import (
    NetqAgentStreamServiceServicer,
    add_NetqAgentStreamServiceServicer_to_server
)
from netq_lib.orm import HOSTNAME

PARTIAL_TRANSFORM = 1

class NetQAgentServicer(NetqAgentStreamServiceServicer):

    run_id = 0
    stop_rpc = False

    def __init__(self, logger, pq, bp_pq):
        self._logger = logger
        self.seq_no = 0
        self._pq = pq
        self._bp_pq = bp_pq
        self.stop_rpc = False
        self.running_rpc_cnt = 0

    def service_start_request(self):
        self._start = True

    def get_lldp_command_out(self):
        msg = ""
        return msg

    def get_sequence_number(self):
        self.seq_no = self.seq_no + 1
        return self.seq_no

    def message_builder(self, msg_type, json_out):

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=self.get_sequence_number(),
                                        ts=long(time.time() * 1000),
                                        trans_mode=PARTIAL_TRANSFORM,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message

    def StreamMessage(self, request, context):
        # pq = NetQPriorityQueue.PQ  # Wrap as utility function
        pq = self._pq
        fd_set = pq.get_reader_fds()
        my_run_id = random.randint(1, 1000)
        peername = 'unknown'
        try:
            peername = context.peer()
        except:
            pass
        self._logger.info("Enable new GRPC %d peer %s pending Q %d", my_run_id, peername, pq.len(1))
        cnt = 0
        if self.running_rpc_cnt:
            self._logger.info("Detected existing GRPC %d", self.run_id)
            # terminate the other rpc
            # indicate to agent thread to go through agent pause resume
            bp_pq = self._bp_pq
            try:
                q_elem = QueueElem(0, 'bp', 'bp')
                bp_pq.putq(q_elem)
            except Exception as ex:
                self._logger.info('GRPC backpressure putq failed %s', ex)
                pass

            self.stop_rpc = True
            while cnt < 3:
                if self.stop_rpc:
                    time.sleep(10)
                else:
                    self.running_rpc_cnt = 0
                    break
                cnt = cnt + 1
        self.running_rpc_cnt = 1
        self.run_id = my_run_id
        log_once = False
        self.stop_rpc = False
        while True:
            if self.stop_rpc:
                self._logger.info("GRPC stream terminated %d", my_run_id)
                self.stop_rpc = False
                return
            try:
                (epoll_input, [], []) = select.select(fd_set, [], [], 10)
                if not len(epoll_input):
                    continue
                if not log_once:
                    self._logger.info("GRPC stream started %d", my_run_id)
                    log_once = True
                for queue_idx in range(0, self._pq.max_num_queue):
                    msg_dict_array = {}
                    for priority, message_type, msg in pq.get_buffer(queue_idx):
                        #json_obj = json.loads(json.dumps(
                        #    msg.data).decode('string-escape').strip('"'))
                        #message_type = json_obj["messageType"]
                        if message_type not in msg_dict_array:
                            msg_dict_array[message_type] = []
                        if isinstance(msg.data, list):
                            msg_dict_array[message_type] = msg.data
                        else:
                            msg_dict_array[message_type].append(msg.data)
                    for message_type, msg_array in msg_dict_array.iteritems():
                        rpc_msg = self.message_builder(message_type, msg_array)
                        yield rpc_msg
            except Exception as e:
                self._logger.error("GRPC %d StreamMessage failed %s", my_run_id, e)
                self.run_id = 0
                self.running_rpc_cnt = 0
                return


class NetQAgentStream(object):

    def __init__(self, conf, logger, pq, bp_pq):
        self._conf = conf
        self._channel = None
        self._logger = logger
        self._pq = pq
        self.run_id = 0
        self._bp_pq = bp_pq
        self.discovery_agent = None

    def _get_server_addr(self):
        # always listen on *:port
        agent_url = "0.0.0.0"
        try:
            agent_port = self._conf['netq-agent'].get(ConfKw.AGENT_PORT, ConfKw.DEFAULT_AGENT_PORT)
        except Exception:
            agent_port = ConfKw.DEFAULT_AGENT_PORT
        server_addr = "{}:{}".format(agent_url, agent_port)
        return server_addr

    def _init_grpc_server(self):
        if getattr(self, 'server', None) is not None:
            self.server.stop(None)
        self.server = None
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        add_NetqAgentStreamServiceServicer_to_server(
            NetQAgentServicer(self._logger, self._pq, self._bp_pq), self.server)

        server_addr = self._get_server_addr()
        self._logger.info("Start Streaming server at {}".format(server_addr))

        self.server.add_insecure_port(server_addr)
        self.server.start()
        self._logger.info("Streaming Server Started")


    def _init_rpc_service(self):

        self._init_grpc_server()

        self.discovery_agent = NetQAgentDiscovery(
            self._conf, logger=self._logger)

        if self.discovery_agent is None:
            self._logger.error("Failed to start disovery process")
            self.server.stop(None)
            return AGENTSTATUS.FAILURE

        discovery_enabled = self._is_discovery_enabled()
        self.wait_time = 10
        self.discovery_retry_time = 60
        retry_count = 0
        prev_wait_time = time.time()
        period_cnt = 0
        while True:
            curr_time = time.time()
            if discovery_enabled:
                if self.get_discovery_status() == ChannelState.IDLE:
                    if not retry_count:
                        self._logger.info("OPTA Start Discovery IDLE")
                    elif retry_count < 3:
                        self._logger.info("OPTA Re-Start Discovery (retry %s) IDLE", retry_count)
                    self.discovery_agent.close()
                    ret = self.discovery_agent.start_discovery(verbose=True)
                    if ret == AGENTSTATUS.FAILURE:
                        # big hammer - start/stop the GRPC server
                        self._logger.info("OPTA Discovery failure - stop/start streaming server")
                        self._init_grpc_server()
                        self._logger.info("OPTA Restarted GRPC server")
                        self.set_discovery_status(ChannelState.IDLE)
                    retry_count += 1
                    # reset the periodic discovery logic
                    prev_wait_time = curr_time
                elif self.get_discovery_status() == ChannelState.CONNECTED:
                    if retry_count:
                        self._logger.info("OPTA Discovery OK READY")
                        retry_count = 0
                    if (curr_time - prev_wait_time) >= self.discovery_retry_time:
                        # calculate next retry time
                        prev_wait_time = curr_time
                        verbose = False
                        if period_cnt == 10:
                            self._logger.info("OPTA Periodic Discovery")
                            verbose = True
                            period_cnt = 0
                        period_cnt += 1
                        self.discovery_agent.close()
                        self.discovery_agent.start_discovery(verbose=verbose)
                        if ret == AGENTSTATUS.FAILURE:
                            # big hammer - start/stop the GRPC server
                            self._logger.info("OPTA Discovery failure - stop/start streaming server")
                            self._init_grpc_server()
                            self._logger.info("OPTA Restarted GRPC server")
                            self.set_discovery_status(ChannelState.IDLE)
                else:
                    # reset the periodic discovery logic
                    prev_wait_time = curr_time
                    if not retry_count:
                        self._logger.info("OPTA Discovery Channel not ready")
                        retry_count += 1
            time.sleep(self.wait_time)

    def start_stream_server(self):
        pass

    def start_stream(self):
        pass

    def stop_stream(self):
        pass

    def set_discovery_status(self, state):
        self.discovery_agent.channel_state = state

    def get_discovery_status(self):
        return self.discovery_agent.channel_state

    def _is_discovery_enabled(self):
        try:
            bool_obj = self._conf['netq-agent'].get(ConfKw.ENABLE_OPTA_DISCOVERY, ConfKw.DEFAULT_ENABLE_OPTA_DISCOVERY)
            if isinstance(bool_obj, bool):
                if bool_obj:
                    return True
                else:
                    return False
        except Exception:
            return True

        if bool_obj.lower() in ['yes', 'true', '1']:
            return True
        else:
            return False

    def stream_main(self):

        self._init_rpc_service()
        return
