# Copyright 2017, 2018 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import OspfNbr, OspfIf
from netq_lib.common.enums import ServiceStatus


def process_ospf_if(cmd_output, svc_failed=False, **kwargs):
    """Process OSPF interface info
    """

    retcode = ServiceStatus.OK

    try:
        json_out = json.loads(cmd_output)
    except ValueError:
        # We treat invalid JSON as the infomation terminating
        json_out = []
        retcode = ServiceStatus.FAIL

    # check if new version of json
    if 'interfaces' in json_out:
        json_out = json_out['interfaces']

    # With OSPF, failed sessions just don't show up. So catch those
    m_ospfif = OspfIf.memmodel()
    for entry in list(m_ospfif.query.filter(hostname=HOSTNAME)):
        new_entry = OspfIf()
        key = entry.ifname
        if key not in json_out or json_out[key].get('area') != entry.area:
            # A key parameter changed or i/f is no longer enabled for ospf
            new_entry.copy(entry)
            new_entry.delete()

    # Handle new entries
    for key in json_out:
        try:
            # ospf json notorious for missing fields - just protect entire block
            entry = OspfIf()
            entry.ifname = key.encode('ascii', 'ignore')
            entry.router_id = json_out[key]['routerId'].encode('ascii', 'ignore')
            entry.area = json_out[key]['area'].encode('ascii', 'ignore')
            entry.if_up = json_out[key].get('ifUp', False)
            entry.is_unnumbered = json_out[key].get('ifUnnumbered', False)
            entry.network_type = json_out[key]['networkType'].encode('ascii',
                                                                 'ignore')
            entry.nbr_count = json_out[key]['nbrCount']
            entry.nbr_adj_count = json_out[key]['nbrAdjacentCount']
            entry.cost = json_out[key]['cost']
            entry.is_passive = json_out[key].get('timerPassiveIface', False)
            entry.mtu = json_out[key]['mtuBytes']
            # Bug in code since time immemorial; the time supplied in JSON
            # isn't in msecs, but its 1000/real_value. So, 5s retransmit time
            # becomes 1000/5 = 200.
            # This is true in version 3.3. In future we'll need to check the
            # version of the OS this agent is running on
            entry.dead_time = 1000 / json_out[key]['timerDeadMsecs']
            entry.rexmit_time = 1000 / json_out[key]['timerRetransmit']
            entry.hello_time = 1000 / json_out[key]['timerMsecs']
            # Can't really check status with this output, need to switch
            entry.save()
        except Exception:
            pass

    return retcode


def process_ospf_neighbors(cmd_output, svc_failed=False, **kwargs):
    """Process OSPF neighbors info
    """

    retcode = ServiceStatus.OK

    try:
        json_out = json.loads(cmd_output)
    except ValueError:
        # We treat invalid JSON as the infomation terminating
        json_out = []
        retcode = ServiceStatus.FAIL

    # With OSPF, failed sessions just don't show up. So catch those
    m_ospfnbr = OspfNbr.memmodel()
    for entry in list(m_ospfnbr.query.filter(hostname=HOSTNAME)):
        if entry.state == 'Fail':
            continue
        new_entry = OspfNbr()
        key = entry.peer_id
        if key not in json_out:
            if svc_failed:
                new_entry.copy(entry)
                new_entry.delete()
            else:
                new_entry.state = 'Fail'
                new_entry.ifname = entry.ifname
                new_entry.area = entry.area
                new_entry.peer_id = entry.peer_id
                new_entry.save()
            retcode = ServiceStatus.ERROR
        elif json_out[key]['areaId'] != entry.area:
            new_entry.copy(entry)
            new_entry.delete()

    # Handle new entries
    for key in json_out:
        try:
            # ospf json notorious for missing fields - just protect entire block
            entry = OspfNbr()
            entry.peer_id = key.encode('ascii', 'ignore')
            entry.peer_addr = json_out[key]['ifaceAddress'].encode('ascii', 'ignore')
            entry.state = json_out[key]['nbrState'].encode('ascii', 'ignore')
            entry.ifname = json_out[key]['ifaceName'].encode('ascii',
                                                         'ignore').split(':')[0]
            entry.area = json_out[key]['areaId'].encode('ascii',
                                                    'ignore').split(':')[0]
            # Can't really check status with this output, need to switch
            entry.save()
        except Exception:
            pass

    return retcode
